# Copyright (c) 2005-2014 Red Hat, Inc.
#
#    Koji is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation;
#    version 2.1 of the License.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this software; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#       Mike McLean <mikem@redhat.com>

# This module provides a threadlocal instance that kojihub uses to store
# request context.
# In the past, we had a custom ThreadLocal implementation, but now this is
# just a thin wrapper around threading.local

from __future__ import absolute_import

import threading


class ThreadLocal(threading.local):
    """A small compatibility wrapper around threading.local"""

    def _threadclear(self):
        self.__dict__.clear()


context = ThreadLocal()



from optparse import OptionParser

from koji import canonArch

from koji.plugin import export_cli
from koji_cli.lib import (
    _running_in_bg,
    activate_session,
    watch_tasks,
    warn,
)


@export_cli
def handle_kiwi_build(goptions, session, args):
    "[build] Run a command in a buildroot"
    usage = "usage: %prog kiwi-build [options] <target> <description_scm> <description_path>"
    usage += "\n(Specify the --help global option for a list of other help options)"
    parser = OptionParser(usage=usage)
    parser.add_option("--scratch", action="store_true", default=False,
                      help="Perform a scratch build")
    parser.add_option("--version", help="Override default version of the output image")
    parser.add_option("--release", help="Release of the output image")
    parser.add_option("--repo", action="append",
                      help="Specify a repo that will override the repo used to install "
                           "RPMs in the image. May be used multiple times. The "
                           "build tag repo associated with the target is the default.")
    parser.add_option("--repo-releasever", help="Override default releasever of the output image")
    parser.add_option("--noprogress", action="store_true",
                      help="Do not display progress of the upload")
    parser.add_option("--kiwi-profile", action="store", default=None,
                      help="Select profile from description file")
    parser.add_option("--type", help="Override default build type from description")
    parser.add_option("--type-attr", action="append", default=[],
                      help="Override default attributes for the build type from description. "
                           "May be used multiple times.")
    parser.add_option("--result-bundle-name-format", help="Override default bundle name format")
    parser.add_option("--make-prep", action="store_true", default=False,
                      help="Run 'make prep' in checkout before starting the build")
    parser.add_option("--buildroot-repo", action="store_true",
                      dest="use_buildroot_repo", default=False,
                      help="Add buildroot repo to installation sources. This is off by default, "
                           "but if there is no --repo used, it will be turned on automatically.")
    parser.add_option("--can-fail", action="store", dest="optional_arches",
                      metavar="ARCH1,ARCH2,...", default="",
                      help="List of archs which are not blocking for build "
                           "(separated by commas.")
    parser.add_option("--arch", action="append", dest="arches", default=[],
                      help="Limit arches to this subset")
    parser.add_option("--nowait", action="store_false", dest="wait", default=True)
    parser.add_option("--wait", action="store_true",
                      help="Wait on the image creation, even if running in the background")
    (options, args) = parser.parse_args(args)

    if len(args) != 3:
        parser.error("Incorrect number of arguments")
        assert False  # pragma: no cover
    target, scm, path = args

    activate_session(session, goptions)
    kwargs = {
        'arches': [],
        'target': target,
        'desc_url': scm,
        'desc_path': path,
    }
    if options.scratch:
        kwargs['scratch'] = True
    if options.optional_arches:
        kwargs['optional_arches'] = [
            canonArch(arch)
            for arch in options.optional_arches.split(',')
            if arch]
    if options.kiwi_profile:
        kwargs['profile'] = options.kiwi_profile
    if options.version:
        kwargs['version'] = options.version
    if options.release:
        kwargs['release'] = options.release
    if options.make_prep:
        kwargs['make_prep'] = True
    if options.type:
        kwargs['type'] = options.type
    if options.type_attr:
        kwargs['type_attr'] = options.type_attr
    if options.result_bundle_name_format:
        kwargs['result_bundle_name_format'] = options.result_bundle_name_format
    if options.arches:
        kwargs['arches'] = [canonArch(arch) for arch in options.arches]
    if options.repo:
        kwargs['repos'] = options.repo
    if options.repo_releasever:
        kwargs['repo_releasever'] = options.repo_releasever

    if session.hub_version < (1, 35, 0):
        warn("hub version is < 1.35, buildroot repo is always used in addition to specified repos")
    elif options.use_buildroot_repo:
        kwargs['use_buildroot_repo'] = True
    elif not options.repo:
        warn("no repos given, using buildroot repo")
        kwargs['use_buildroot_repo'] = True

    task_id = session.kiwiBuild(**kwargs)

    if not goptions.quiet:
        print("Created task: %d" % task_id)
        print("Task info: %s/taskinfo?taskID=%s" % (goptions.weburl, task_id))
    if options.wait or (options.wait is None and not _running_in_bg()):
        session.logout()
        return watch_tasks(session, [task_id], quiet=goptions.quiet,
                           poll_interval=goptions.poll_interval, topurl=goptions.topurl)



from __future__ import absolute_import
from unittest import mock
import shutil
import six
import tempfile
import unittest

import koji
import koji.daemon
import koji.policy

from koji.daemon import SCM

policy = {
    'one': '''
        match scm_host goodserver :: allow none
        match scm_host badserver :: deny
        match scm_host maybeserver && match scm_repository /badpath/* :: deny
        all :: allow
    ''',
    'two': '''
        match scm_host default :: allow
        match scm_host nocommon :: allow
        match scm_host common :: allow use_common
        match scm_host srccmd :: allow fedpkg sources
        match scm_host nosrc :: allow none
        match scm_host mixed && match scm_repository /foo/* :: allow
        match scm_host mixed && match scm_repository /bar/* :: allow use_common
        match scm_host mixed && match scm_repository /baz/* :: allow fedpkg sources
        match scm_host mixed && match scm_repository /foobar/* :: allow use_common fedpkg sources
        match scm_host mixed && match scm_repository /foobaz/* :: allow use_common none
    '''
}


class FakePolicy(object):

    def __init__(self, rule):
        base_tests = koji.policy.findSimpleTests(vars(koji.policy))
        self.ruleset = koji.policy.SimpleRuleSet(rule.splitlines(), base_tests)

    def evalPolicy(self, name, data):
        return self.ruleset.apply(data)


class TestSCM(unittest.TestCase):

    def test_urlcheck(self):
        good = [
            "git://server/foo.git#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "git+ssh://server2/other/path#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "svn://server/path/to/code#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "svn+ssh://server/some/path#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "cvs://server/some/path#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "cvs+ssh://server/some/path#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "cvs+ssh://server:80/some/path#bab0c73900241ef5c465d7e873e9d8b34c948e67",
        ]
        bad = [
            "http://localhost/foo.html",
            "foo-1.1-1.src.rpm",
            "https://server/foo-1.1-1.src.rpm",
            "git:foobar",
            "https:foo/bar",
            "https://",
        ]
        for url in good:
            self.assertTrue(SCM.is_scm_url(url))
        for url in bad:
            self.assertFalse(SCM.is_scm_url(url))

    @mock.patch('logging.getLogger')
    def test_init(self, getLogger):
        bad = [
            "git://user@@server/foo.git#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "git://user:pass@server/foo.git#bab0c73900241ef5c465d7e873e9d8b34c948e67",
            "git://server/foo.git;params=not_allowed",
            "git://server#asdasd",  # no path
            "git://server/foo.git",  # no fragment
            "http://localhost/foo.html",
            "git://@localhost/foo/?a=bar/",
            "http://localhost/foo.html?a=foo/",
            "foo-1.1-1.src.rpm",
            "git://",
            "https://server/foo-1.1-1.src.rpm",
        ]
        for url in bad:
            with self.assertRaises(koji.GenericError):
                scm = SCM(url)

        url = "git://user@server/foo.git#bab0c73900241ef5c465d7e873e9d8b34c948e67"
        scm = SCM(url)
        self.assertEqual(scm.scheme, 'git://')
        self.assertEqual(scm.user, 'user')
        self.assertEqual(scm.host, 'server')
        self.assertEqual(scm.repository, '/foo.git')
        self.assertEqual(scm.module, '')
        self.assertEqual(scm.revision, 'bab0c73900241ef5c465d7e873e9d8b34c948e67')
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])
        self.assertEqual(scm.scmtype, 'GIT')

    def test_with_port(self):
        url = "git+ssh://user@server:80/foo.git#bab0c73900241ef5c465d7e873e9d8b34c948e67"
        scm = SCM(url)
        self.assertEqual(scm.scheme, 'git+ssh://')
        self.assertEqual(scm.user, 'user')
        self.assertEqual(scm.host, 'server:80')
        self.assertEqual(scm.repository, '/foo.git')
        self.assertEqual(scm.module, '')
        self.assertEqual(scm.revision, 'bab0c73900241ef5c465d7e873e9d8b34c948e67')
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])
        self.assertEqual(scm.scmtype, 'GIT+SSH')

    def test_assert_allowed_basic(self):
        scm = SCM("git://scm.example.com/path1#1234")

        # session must be passed
        with self.assertRaises(koji.GenericError) as cm:
            scm.assert_allowed(session=None, by_config=False, by_policy=True)
        self.assertEqual(str(cm.exception),
                         'When allowed SCM assertion is by policy, session must be passed in.')

        # allowed could not be None
        scm.assert_allowed_by_config = mock.MagicMock()
        scm.assert_allowed(allowed=None, by_config=True, by_policy=False)
        scm.assert_allowed_by_config.assert_called_once_with('')

    @mock.patch('logging.getLogger')
    def test_allowed_by_config(self, getLogger):
        config = '''
            goodserver:*:no
            !badserver:*
            !maybeserver:/badpath/*
            maybeserver:*:no
            '''
        good = [
            "git://goodserver/path1#1234",
            "git+ssh://maybeserver/path1#1234",
        ]
        bad = [
            "cvs://badserver/projects/42#ref",
            "svn://badserver/projects/42#ref",
            "git://maybeserver/badpath/project#1234",
            "git://maybeserver//badpath/project#1234",
            "git://maybeserver////badpath/project#1234",
            "git://maybeserver/./badpath/project#1234",
            "git://maybeserver//.//badpath/project#1234",
            "git://maybeserver/goodpath/../badpath/project#1234",
            "git://maybeserver/goodpath/..//badpath/project#1234",
            "git://maybeserver/..//badpath/project#1234",
        ]
        for url in good:
            scm = SCM(url)
            scm.assert_allowed(config)
        for url in bad:
            scm = SCM(url)
            try:
                scm.assert_allowed(config)
            except koji.BuildError:
                pass
            else:
                raise AssertionError("allowed bad url: %s" % url)

    @mock.patch('logging.getLogger')
    def test_badrule_by_config(self, getLogger):
        config = '''
            bogus-entry-should-be-ignored
            goodserver:*:no
            !badserver:*
            '''
        url = "git://goodserver/path1#1234"
        scm = SCM(url)
        scm.assert_allowed(config)

    @mock.patch('logging.getLogger')
    def test_opts_by_config(self, getLogger):
        config = '''
            default:*
            nocommon:*:no
            srccmd:*:no:fedpkg,sources
            nosrc:*:no:
            mixed:/foo/*:no
            mixed:/bar/*:yes
            mixed:/baz/*:no:fedpkg,sources
            '''

        url = "git://default/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://nocommon/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://srccmd/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://nosrc/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, None)

        url = "git://mixed/foo/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/bar/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/baz/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://mixed/koji.git#1234"
        scm = SCM(url)
        with self.assertRaises(koji.BuildError):
            scm.assert_allowed(config)

        url = "git://mixed/foo/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/bar/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/baz/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed(config)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://mixed/koji.git#1234"
        scm = SCM(url)
        with self.assertRaises(koji.BuildError):
            scm.assert_allowed(config)

    def test_allowed_by_policy(self):
        good = [
            "git://goodserver/path1#1234",
            "git+ssh://maybeserver/path1#1234",
        ]
        bad = [
            "cvs://badserver/projects/42#ref",
            "svn://badserver/projects/42#ref",
            "git://maybeserver/badpath/project#1234",
            "git://maybeserver//badpath/project#1234",
            "git://maybeserver////badpath/project#1234",
            "git://maybeserver/./badpath/project#1234",
            "git://maybeserver//.//badpath/project#1234",
            "git://maybeserver/goodpath/../badpath/project#1234",
            "git://maybeserver/goodpath/..//badpath/project#1234",
            "git://maybeserver/..//badpath/project#1234",
        ]
        session = mock.MagicMock()
        session.evalPolicy.side_effect = FakePolicy(policy['one']).evalPolicy
        for url in good:
            scm = SCM(url)
            scm.assert_allowed(session=session, by_config=False, by_policy=True)
        for url in bad:
            scm = SCM(url)
            with self.assertRaises(koji.BuildError) as cm:
                scm.assert_allowed(session=session, by_config=False, by_policy=True)
            self.assertRegex(str(cm.exception), '^SCM: .* is not allowed, reason: None$')

    def test_opts_by_policy(self):
        session = mock.MagicMock()
        session.evalPolicy.side_effect = FakePolicy(policy['two']).evalPolicy

        url = "git://default/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://nocommon/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://common/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://srccmd/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://nosrc/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, None)

        url = "git://mixed/foo/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/bar/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/baz/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://mixed/koji.git#1234"
        scm = SCM(url)
        with self.assertRaises(koji.BuildError):
            scm.assert_allowed_by_policy(session=session)

        url = "git://mixed/foo/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/bar/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/baz/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://mixed/foobar/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://mixed/foobaz/koji.git#1234"
        scm = SCM(url)
        scm.assert_allowed_by_policy(session=session)
        self.assertEqual(scm.use_common, True)
        self.assertIsNone(scm.source_cmd)

        url = "git://nomatch/koji.git#1234"
        scm = SCM(url)
        with self.assertRaises(koji.BuildError):
            scm.assert_allowed_by_policy(session=session)

    def test_assert_allowed_by_both(self):
        config = '''
            default:*:no:
            mixed:/foo/*:yes
            mixed:/bar/*:no
            mixed:/baz/*:no:centpkg,sources
            mixed:/foobar/*:no:
            mixed:/foobaz/*:no:centpkg,sources
            '''

        session = mock.MagicMock()
        session.evalPolicy.side_effect = FakePolicy(policy['two']).evalPolicy

        url = "git://default/koji.git#1234"
        scm = SCM(url)
        # match scm_host default :: allow
        scm.assert_allowed(allowed=config, session=session, by_config=True, by_policy=True)
        self.assertEqual(scm.use_common, False)
        self.assertIsNone(scm.source_cmd)

        url = "git://mixed/foo/koji.git#1234"
        scm = SCM(url)
        # match scm_host mixed && match scm_repository /foo/* :: allow
        scm.assert_allowed(allowed=config, session=session, by_config=True, by_policy=True)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/bar/koji.git#1234"
        scm = SCM(url)
        # match scm_host mixed && match scm_repository /bar/* :: allow use_common
        scm.assert_allowed(allowed=config, session=session, by_config=True, by_policy=True)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['make', 'sources'])

        url = "git://mixed/baz/koji.git#1234"
        scm = SCM(url)
        # match scm_host mixed && match scm_repository /baz/* :: allow fedpkg sources
        scm.assert_allowed(allowed=config, session=session, by_config=True, by_policy=True)
        self.assertEqual(scm.use_common, False)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://mixed/foobar/koji.git#1234"
        scm = SCM(url)
        # match scm_host mixed && match scm_repository /foobar/* :: allow use_common fedpkg sources
        scm.assert_allowed(allowed=config, session=session, by_config=True, by_policy=True)
        self.assertEqual(scm.use_common, True)
        self.assertEqual(scm.source_cmd, ['fedpkg', 'sources'])

        url = "git://mixed/foobaz/koji.git#1234"
        scm = SCM(url)
        # match scm_host mixed && match scm_repository /foobaz/* :: allow use_common none
        scm.assert_allowed(allowed=config, session=session, by_config=True, by_policy=True)
        self.assertEqual(scm.use_common, True)
        self.assertIsNone(scm.source_cmd)


class TestSCMCheckouts(unittest.TestCase):

    def setUp(self):
        self.symlink = mock.patch('os.symlink').start()
        self.getLogger = mock.patch('logging.getLogger').start()
        self.log_output = mock.patch('koji.daemon.log_output').start()
        self.log_output.return_value = None
        self.tempdir = tempfile.mkdtemp()
        self.session = mock.MagicMock()
        self.uploadpath = mock.MagicMock()
        self.logfile = '/dev/null'
        self.config = '''
            default:*
            nocommon:*:no
            srccmd:*:no:fedpkg,sources
            nosrc:*:no:
            '''

    def tearDown(self):
        mock.patch.stopall()
        shutil.rmtree(self.tempdir)

    def test_checkout_git_nocommon(self):

        url = "git://nocommon/koji.git#asdasd"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)
        self.assertEqual(scm.use_common, False)
        self.symlink.assert_not_called()
        # expected commands
        cmd = ['git', 'clone', '-n', 'git://nocommon/koji.git', self.tempdir + '/koji']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env=None)
        cmd = ['git', 'fetch', 'origin', 'asdasd:KOJI_FETCH_HEAD']
        call2 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir + '/koji', logerror=1, append=True, env=None)
        cmd = ['git', 'reset', '--hard', 'KOJI_FETCH_HEAD']
        call3 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir + '/koji', logerror=1, append=True, env=None)
        self.log_output.assert_has_calls([call1, call2, call3])

    def test_checkout_gitssh_nocommon(self):

        url = "git+ssh://user@nocommon/koji.git#asdasd"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)
        self.assertEqual(scm.use_common, False)
        self.symlink.assert_not_called()
        # expected commands
        cmd = ['git', 'clone', '-n', 'git+ssh://user@nocommon/koji.git', self.tempdir + '/koji']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env=None)
        cmd = ['git', 'fetch', 'origin', 'asdasd:KOJI_FETCH_HEAD']
        call2 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir + '/koji', logerror=1, append=True, env=None)
        cmd = ['git', 'reset', '--hard', 'KOJI_FETCH_HEAD']
        call3 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir + '/koji', logerror=1, append=True, env=None)
        self.log_output.assert_has_calls([call1, call2, call3])

    def test_checkout_git_common(self):

        url = "git://default/koji.git#asdasd"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session, uploadpath=self.uploadpath,
                     logfile=self.logfile)
        self.assertEqual(scm.use_common, True)
        self.symlink.assert_called_once()
        # expected commands
        cmd = ['git', 'clone', '-n', 'git://default/koji.git', self.tempdir + '/koji']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env=None)
        cmd = ['git', 'fetch', 'origin', 'asdasd:KOJI_FETCH_HEAD']
        call2 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir + '/koji', logerror=1, append=True, env=None)
        cmd = ['git', 'reset', '--hard', 'KOJI_FETCH_HEAD']
        call3 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir + '/koji', logerror=1, append=True, env=None)
        cmd = ['git', 'clone', 'git://default/common.git', 'common']
        call4 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=True, env=None)
        self.log_output.assert_has_calls([call1, call2, call3, call4])

    def test_checkout_error_in_command(self):

        url = "git://nocommon/koji.git#asdasd"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        self.log_output.return_value = 1
        with self.assertRaises(koji.BuildError):
            scm.checkout(self.tempdir, session=self.session,
                         uploadpath=self.uploadpath, logfile=self.logfile)
        self.assertEqual(scm.use_common, False)
        self.symlink.assert_not_called()
        # expected commands
        cmd = ['git', 'clone', '-n', 'git://nocommon/koji.git', self.tempdir + '/koji']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env=None)
        # should have errored after first command
        self.log_output.assert_has_calls([call1])

    def test_checkout_cvs_common(self):

        url = "cvs://default/cvsisdead?rpms/foo/EL3#sometag"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)
        self.assertEqual(scm.use_common, True)
        self.symlink.assert_called_once()
        # expected commands
        cmd = ['cvs', '-d', ':pserver:anonymous@default:/cvsisdead', 'checkout',
               '-r', 'sometag', 'rpms/foo/EL3']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env=None)
        cmd = ['cvs', '-d', ':pserver:anonymous@default:/cvsisdead', 'checkout', 'common']
        call2 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=True, env=None)
        self.log_output.assert_has_calls([call1, call2])

    def test_checkout_cvs_ssh(self):

        url = "cvs+ssh://user@nocommon/cvsisdead?rpms/foo/EL3#sometag"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)
        self.assertEqual(scm.use_common, False)
        self.symlink.assert_not_called()
        # expected commands
        cmd = ['cvs', '-d', ':ext:user@nocommon:/cvsisdead', 'checkout', '-r',
               'sometag', 'rpms/foo/EL3']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env={'CVS_RSH': 'ssh'})
        self.log_output.assert_has_calls([call1])

    def test_checkout_svn(self):

        url = "svn://nocommon/dist?rpms/foo/EL3#revision"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)
        self.assertEqual(scm.use_common, False)
        self.symlink.assert_not_called()
        # expected commands
        cmd = ['svn', 'checkout', '-r', 'revision',
               'svn://nocommon/dist/rpms/foo/EL3', 'rpms/foo/EL3']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env=None)
        self.log_output.assert_has_calls([call1])

    def test_checkout_svn_ssh(self):

        url = "svn+ssh://user@nocommon/dist?rpms/foo/EL3#revision"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)
        self.assertEqual(scm.use_common, False)
        self.symlink.assert_not_called()
        # expected commands
        cmd = ['svn', 'checkout', '-r', 'revision',
               'svn+ssh://user@nocommon/dist/rpms/foo/EL3', 'rpms/foo/EL3']
        call1 = mock.call(self.session, cmd[0], cmd, self.logfile, self.uploadpath,
                          cwd=self.tempdir, logerror=1, append=False, env=None)
        self.log_output.assert_has_calls([call1])

    @mock.patch('subprocess.Popen')
    def test_get_source_git(self, popen):
        popen.return_value.wait.return_value = 0
        popen.return_value.communicate = mock.MagicMock()
        popen.return_value.communicate.return_value = (six.b('hash '), six.b('any'))

        url = "git://default/koji.git#asdasd"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)

        source = scm.get_source()
        self.assertEqual(source, {'url': url, 'source': 'git://default/koji.git#hash'})

        popen.return_value.wait.return_value = 1
        with self.assertRaises(koji.GenericError) as cm:
            source = scm.get_source()
        self.assertEqual(cm.exception.args[0],
                         'Error getting commit hash for git')

    @mock.patch('subprocess.Popen')
    def test_get_source_other(self, popen):
        popen.return_value.wait.return_value = 0
        popen.return_value.communicate = mock.MagicMock()
        popen.return_value.communicate.return_value = ('hash ', 'any')

        url = "svn+ssh://user@nocommon/dist?rpms/foo/EL3#revision"
        scm = SCM(url)
        scm.assert_allowed(self.config)
        scm.checkout(self.tempdir, session=self.session,
                     uploadpath=self.uploadpath, logfile=self.logfile)

        source = scm.get_source()
        self.assertEqual(source, {'url': url, 'source': url})



from __future__ import absolute_import
import os
import six
import subprocess
import unittest
import sys

try:
    import importlib.util
except ImportError:  # pragma: no cover
    import imp as imp
    importlib = None

# docs version lives in docs/source/conf.py
TOPDIR = os.path.dirname(__file__) + '/..'
SPHINX_CONF = TOPDIR + '/docs/source/conf.py'

if importlib:
    spec = importlib.util.spec_from_file_location("sphinx_conf", SPHINX_CONF)
    sphinx_conf = importlib.util.module_from_spec(spec)
    sys.modules["sphinx_conf"] = sphinx_conf
    spec.loader.exec_module(sphinx_conf)
else:
    sphinx_conf = imp.load_source('sphinx_conf', SPHINX_CONF)


class TestDocsVersion(unittest.TestCase):

    def get_spec(self):
        return TOPDIR + '/koji.spec'

    def get_koji_version(self):
        spec = self.get_spec()
        cmd = ['rpm', '-q', '--specfile', spec, '--qf', '%{version}\\n']
        popen = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        output = popen.stdout.read()
        # rpm outputs a line for each subpackage
        version = output.splitlines()[0]
        if six.PY3:
            version = version.decode()
        return version

    def test_docs_version(self):
        koji_version = self.get_koji_version()
        self.assertEqual(koji_version, sphinx_conf.release)
        # docs 'version' is x.y instead of x.y.z
        dver = '.'.join(koji_version.split('.')[:-1])
        self.assertEqual(dver, sphinx_conf.version)



# coding: utf-8
# Copyright © 2019 Red Hat, Inc.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import absolute_import

from optparse import OptionParser

import koji
from koji.plugin import export_cli
from koji_cli.commands import anon_handle_wait_repo
from koji_cli.lib import activate_session, arg_filter


@export_cli
def handle_add_sidetag(options, session, args):
    "Create sidetag"
    usage = "%prog add-sidetag [options] <basetag>"
    usage += "\n(Specify the --help global option for a list of other help options)"
    parser = OptionParser(usage=usage)
    parser.add_option("-q", "--quiet", action="store_true", help="Do not print tag name",
                      default=options.quiet)
    parser.add_option("-w", "--wait", action="store_true", help="Wait until repo is ready.")
    parser.add_option("--debuginfo", action="store_true",
                      help="Buildroot repo will contain debuginfos")
    parser.add_option("--suffix", action="store", help="Suffix from hub-supported ones")

    (opts, args) = parser.parse_args(args)

    if len(args) != 1:
        parser.error("Only argument is basetag")
    basetag = args[0]

    activate_session(session, options)

    kwargs = {"debuginfo": opts.debuginfo}
    if opts.suffix:
        kwargs['suffix'] = opts.suffix
    try:
        tag = session.createSideTag(basetag, **kwargs)
    except koji.ActionNotAllowed:
        parser.error("Policy violation")
    except koji.ParameterError as ex:
        if 'suffix' in str(ex):
            parser.error("Hub is older and doesn't support --suffix, please run it without it")
        else:
            raise

    if not opts.quiet:
        print(tag["name"])

    if opts.wait:
        args = ["--target", tag["name"]]
        if opts.quiet:
            args.append("--quiet")
        anon_handle_wait_repo(options, session, args)


@export_cli
def handle_remove_sidetag(options, session, args):
    "Remove sidetag"
    usage = "%prog remove-sidetag [options] <sidetag> ..."
    usage += "\n(Specify the --help global option for a list of other help options)"
    parser = OptionParser(usage=usage)
    (opts, args) = parser.parse_args(args)

    if len(args) < 1:
        parser.error("Sidetag argument is required")

    activate_session(session, options)

    session.multicall = True
    for sidetag in args:
        session.removeSideTag(sidetag)
    session.multiCall(strict=True)


@export_cli
def handle_list_sidetags(options, session, args):
    "List sidetags"
    usage = "%prog list-sidetags [options]"
    usage += "\n(Specify the --help global option for a list of other help options)"
    parser = OptionParser(usage=usage)
    parser.add_option("--basetag", action="store", help="Filter on basetag")
    parser.add_option("--user", action="store", help="Filter on user")
    parser.add_option("--mine", action="store_true", help="Filter on user")

    (opts, args) = parser.parse_args(args)

    if len(args) > 0:
        parser.error("This command takes no arguments")

    if opts.mine and opts.user:
        parser.error("Specify only one from --user --mine")

    if opts.mine:
        activate_session(session, options)
        user = session.getLoggedInUser()["name"]
    else:
        user = opts.user

    for tag in session.listSideTags(basetag=opts.basetag, user=user):
        print(tag["name"])


@export_cli
def handle_edit_sidetag(options, session, args):
    "Edit sidetag"
    usage = "%prog edit-sidetag [options] <sidetag>"
    usage += "\n(Specify the --help global option for a list of other help options)"
    parser = OptionParser(usage=usage)
    parser.add_option("--debuginfo", action="store_true", default=None,
                      help="Generate debuginfo repository")
    parser.add_option("--no-debuginfo", action="store_false", dest="debuginfo")
    parser.add_option("--rpm-macro", action="append", default=[], metavar="key=value",
                      dest="rpm_macros", help="Set tag-specific rpm macros")
    parser.add_option("--remove-rpm-macro", action="append", default=[], metavar="key",
                      dest="remove_rpm_macros", help="Remove rpm macros")

    (opts, args) = parser.parse_args(args)

    if len(args) != 1:
        parser.error("Only argument is sidetag")
    sidetag = args[0]

    if opts.debuginfo is None and not opts.rpm_macros and not opts.remove_rpm_macros:
        parser.error("At least one option needs to be specified")

    activate_session(session, options)

    kwargs = {}
    if opts.debuginfo is not None:
        kwargs['debuginfo'] = opts.debuginfo

    if opts.rpm_macros:
        rpm_macros = {}
        for xopt in opts.rpm_macros:
            key, value = xopt.split('=', 1)
            value = arg_filter(value)
            rpm_macros[key] = value
        kwargs['rpm_macros'] = rpm_macros

    if opts.remove_rpm_macros:
        kwargs['remove_rpm_macros'] = opts.remove_rpm_macros

    session.editSideTag(sidetag, **kwargs)



# Copyright (C) 2006 Mandriva; 2009-2014 Red Hat, Inc.
# Authors: Frederic Lepied, Florian Festi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
#
# This library and program is heavily based on rpmdiff from the rpmlint package
# It was modified to be used as standalone library for the Koji project.

from __future__ import absolute_import

import hashlib
import itertools
import json
import os

import rpm
import six
from six.moves import zip


class BytesJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if six.PY3 and isinstance(o, bytes):
            return o.decode('utf-8')
        return json.JSONEncoder.default(self, o)


class Rpmdiff:

    # constants

    TAGS = (rpm.RPMTAG_NAME, rpm.RPMTAG_SUMMARY,
            rpm.RPMTAG_DESCRIPTION, rpm.RPMTAG_GROUP,
            rpm.RPMTAG_LICENSE, rpm.RPMTAG_URL,
            rpm.RPMTAG_PREIN, rpm.RPMTAG_POSTIN,
            rpm.RPMTAG_PREUN, rpm.RPMTAG_POSTUN)

    PRCO = ('REQUIRES', 'PROVIDES', 'CONFLICTS', 'OBSOLETES')

    # {fname : (size, mode, mtime, flags, dev, inode,
    #          nlink, state, vflags, user, group, digest)}
    __FILEIDX = [['S', 0],
                 ['M', 1],
                 ['5', 11],
                 ['D', 4],
                 ['N', 6],
                 ['L', 7],
                 ['V', 8],
                 ['U', 9],
                 ['G', 10],
                 ['F', 3],
                 ['T', 2]]

    try:
        if rpm.RPMSENSE_SCRIPT_PRE:
            PREREQ_FLAG = rpm.RPMSENSE_PREREQ | rpm.RPMSENSE_SCRIPT_PRE |\
                rpm.RPMSENSE_SCRIPT_POST | rpm.RPMSENSE_SCRIPT_PREUN |\
                rpm.RPMSENSE_SCRIPT_POSTUN
    except AttributeError:
        try:
            PREREQ_FLAG = rpm.RPMSENSE_PREREQ
        except Exception:
            # (proyvind): This seems ugly, but then again so does
            #            this whole check as well.
            PREREQ_FLAG = False

    DEPFORMAT = '%-12s%s %s %s %s'
    FORMAT = '%-12s%s'

    ADDED = 'added'
    REMOVED = 'removed'

    # code starts here

    def __init__(self, old, new, ignore=None):
        self.result = []
        self.old_data = {'tags': {}, 'ignore': ignore}
        self.new_data = {'tags': {}, 'ignore': ignore}
        if ignore is None:
            ignore = set()
        else:
            ignore = set(ignore)

        old = self.__load_pkg(old)
        new = self.__load_pkg(new)

        # Compare single tags
        for tag in self.TAGS:
            old_tag = old[tag]
            new_tag = new[tag]
            self.old_data['tags'][tag] = old[tag]
            self.new_data['tags'][tag] = new[tag]
            if old_tag != new_tag:
                tagname = rpm.tagnames[tag]
                if old_tag is None:
                    self.__add(self.FORMAT, (self.ADDED, tagname))
                elif new_tag is None:
                    self.__add(self.FORMAT, (self.REMOVED, tagname))
                else:
                    self.__add(self.FORMAT, ('S.5........', tagname))

        # compare Provides, Requires, ...
        for tag in self.PRCO:
            self.__comparePRCOs(old, new, tag)

        # compare the files

        old_files_dict = self.__getFilesDict(old)
        new_files_dict = self.__getFilesDict(new)

        files = sorted(set(itertools.chain(six.iterkeys(old_files_dict),
                                           six.iterkeys(new_files_dict))))
        self.old_data['files'] = old_files_dict
        self.new_data['files'] = new_files_dict

        for f in files:
            diff = 0

            old_file = old_files_dict.get(f)
            new_file = new_files_dict.get(f)

            if not old_file:
                self.__add(self.FORMAT, (self.ADDED, f))
            elif not new_file:
                self.__add(self.FORMAT, (self.REMOVED, f))
            else:
                format = ''
                for entry in self.__FILEIDX:
                    # entry = [character, value]
                    if entry[0] in ignore:
                        # erase fields which are ignored
                        old_file[entry[1]] = None
                        new_file[entry[1]] = None
                        format = format + '.'
                    elif old_file[entry[1]] != new_file[entry[1]]:
                        format = format + entry[0]
                        diff = 1
                    else:
                        format = format + '.'
                if diff:
                    self.__add(self.FORMAT, (format, f))

    # return a report of the differences
    def textdiff(self):
        return '\n'.join((format % data for format, data in self.result))

    # do the two rpms differ
    def differs(self):
        return bool(self.result)

    # add one differing item
    def __add(self, format, data):
        self.result.append((format, data))

    # load a package from a file or from the installed ones
    def __load_pkg(self, filename):
        ts = rpm.ts()
        f = os.open(filename, os.O_RDONLY)
        hdr = ts.hdrFromFdno(f)
        os.close(f)
        return hdr

    # output the right string according to RPMSENSE_* const
    def sense2str(self, sense):
        s = ""
        for tag, char in ((rpm.RPMSENSE_LESS, "<"),
                          (rpm.RPMSENSE_GREATER, ">"),
                          (rpm.RPMSENSE_EQUAL, "=")):
            if sense & tag:
                s += char
        return s

    # compare Provides, Requires, Conflicts, Obsoletes
    def __comparePRCOs(self, old, new, name):
        oldflags = old[name[:-1] + 'FLAGS']
        newflags = new[name[:-1] + 'FLAGS']
        # fix buggy rpm binding not returning list for single entries
        if not isinstance(oldflags, list):
            oldflags = [oldflags]
        if not isinstance(newflags, list):
            newflags = [newflags]

        o = list(zip(old[name], oldflags, old[name[:-1] + 'VERSION']))
        n = list(zip(new[name], newflags, new[name[:-1] + 'VERSION']))

        if name == 'PROVIDES':  # filter our self provide
            oldNV = (old['name'], rpm.RPMSENSE_EQUAL,
                     "%s-%s" % (old['version'], old['release']))
            newNV = (new['name'], rpm.RPMSENSE_EQUAL,
                     "%s-%s" % (new['version'], new['release']))
            o = [entry for entry in o if entry != oldNV]
            n = [entry for entry in n if entry != newNV]

        self.old_data[name] = sorted(o)
        self.new_data[name] = sorted(n)

        for oldentry in o:
            if oldentry not in n:
                if name == 'REQUIRES' and oldentry[1] & self.PREREQ_FLAG:
                    tagname = 'PREREQ'
                else:
                    tagname = name
                self.__add(self.DEPFORMAT,
                           (self.REMOVED, tagname, oldentry[0],
                            self.sense2str(oldentry[1]), oldentry[2]))
        for newentry in n:
            if newentry not in o:
                if name == 'REQUIRES' and newentry[1] & self.PREREQ_FLAG:
                    tagname = 'PREREQ'
                else:
                    tagname = name
                self.__add(self.DEPFORMAT,
                           (self.ADDED, tagname, newentry[0],
                            self.sense2str(newentry[1]), newentry[2]))

    def __getFilesDict(self, hdr):
        if not hasattr(rpm, 'files'):
            # fall back to file iterator
            return self.__fileIteratorToDict(hdr.fiFromHeader())
        result = {}
        for file in rpm.files(hdr):
            # mimic old fi order because comparison and kojihash relies on it
            # (FN, FSize, FMode, FMtime, FFlags, FRdev, FInode, FNlink, FState, VFlags, FUser,
            # FGroup, Digest)
            filedata = [
                # name omitted
                file.size,
                file.mode,
                file.mtime,
                file.fflags,
                file.rdev,
                file.inode,
                file.nlink,
                file.state,
                file.vflags,
                file.user,
                file.group,
                file.digest,
            ]
            result[file.name] = filedata
        return result

    def __fileIteratorToDict(self, fi):
        result = {}
        for filedata in fi:
            # index by name
            result[filedata[0]] = list(filedata[1:])
        return result

    def kojihash(self, new=False):
        """return hashed data for use in koji"""
        if new:
            data = self.new_data
        else:
            data = self.old_data
        if not data:
            raise ValueError("rpm header data are empty")
        s = json.dumps(data, sort_keys=True, cls=BytesJSONEncoder)
        if six.PY3:
            s = s.encode('utf-8')
        return hashlib.sha256(s).hexdigest()



# Copyright (c) 2005-2014 Red Hat, Inc.
#
#    Koji is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation;
#    version 2.1 of the License.
#
#    This software is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this software; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#       Mike McLean <mikem@redhat.com>
#       Mike Bonnet <mikeb@redhat.com>

from __future__ import absolute_import, division

import base64
import calendar
import datetime
import errno
import hashlib
import json
import logging
import os
import os.path
import re
import resource
import shutil
import stat
import struct
import sys
import tempfile
import time
import warnings
from fnmatch import fnmatch
from zlib import adler32

import six
from six.moves import range, zip

import koji
from koji.xmlrpcplus import DateTime


# BEGIN kojikamid dup #

def md5_constructor(*args, **kwargs):
    if hasattr(hashlib._hashlib, 'get_fips_mode') and hashlib._hashlib.get_fips_mode():
        # do not care about FIPS we need md5 for signatures and older hashes
        # It is still used for *some* security
        kwargs['usedforsecurity'] = False
    return hashlib.md5(*args, **kwargs)  # nosec

# END kojikamid dup #


# imported from kojiweb and kojihub
def deprecated(message):
    """Print deprecation warning"""
    with warnings.catch_warnings():
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(message, DeprecationWarning)


def _changelogDate(cldate):
    return time.strftime('%a %b %d %Y',
                         time.strptime(koji.formatTime(cldate), '%Y-%m-%d %H:%M:%S'))


def formatChangelog(entries):
    """Format a list of changelog entries (dicts)
    into a string representation."""
    result = ''
    for entry in entries:
        result += """* %s %s
%s

""" % (_changelogDate(entry['date']),
            koji._fix_print(entry['author']),
            koji._fix_print(entry['text']))
    return result


DATE_RE = re.compile(r'(\d+)-(\d+)-(\d+)')
TIME_RE = re.compile(r'(\d+):(\d+):(\d+)')


def parseTime(val):
    """
    Parse a string time in either "YYYY-MM-DD HH24:MI:SS" or "YYYY-MM-DD"
    format into floating-point seconds since the epoch.  If the time portion
    is not specified, it will be padded with zeros.  The string time is treated
    as UTC.  If the time string cannot be parsed into a valid date, None will be
    returned.
    """
    result = DATE_RE.search(val)
    if not result:
        return None
    else:
        date = [int(r) for r in result.groups()]
    time = [0, 0, 0]
    rest = val[result.end():].strip()
    result = TIME_RE.search(rest)
    if result:
        time = [int(r) for r in result.groups()]
    return calendar.timegm(
        datetime.datetime(*(date + time)).timetuple())


def checkForBuilds(session, tag, builds, event, latest=False):
    """Check that the builds existed in tag at the time of the event.
       If latest=True, check that the builds are the latest in tag."""
    for build in builds:
        if latest:
            tagged_list = session.getLatestBuilds(tag, event=event, package=build['name'])
        else:
            tagged_list = session.listTagged(tag, event=event, package=build['name'], inherit=True)
        for tagged in tagged_list:
            if tagged['version'] == build['version'] and tagged['release'] == build['release']:
                break
        else:
            return False

    return True


class RepoWatcher(object):

    # timing defaults
    PAUSE = 6
    TIMEOUT = 120

    def __init__(self, session, tag, nvrs=None, min_event=None, at_event=None, opts=None,
                 logger=None):
        self.session = session
        self.taginfo = session.getTag(tag, strict=True)
        self.start = None
        if nvrs is None:
            nvrs = []
        self.nvrs = nvrs
        self.builds = [koji.parse_NVR(nvr) for nvr in nvrs]
        # note that we don't assume the nvrs exist yet
        self.at_event = at_event
        if min_event is None:
            self.min_event = None
        elif at_event is not None:
            raise koji.ParameterError('Cannot specify both min_event and at_event')
        elif min_event == "last":
            self.min_event = session.tagLastChangeEvent(self.taginfo['id'])
        else:
            self.min_event = int(min_event)
        # if opts is None we'll get the default opts
        self.opts = opts
        self.logger = logger or logging.getLogger('koji')
        self._task_request = None

    def get_start(self):
        # we don't want necessarily want to start the clock in init
        if not self.start:
            self.start = time.time()
        return self.start

    def getRepo(self):
        """Return repo if available now, without waiting

        Returns repoinfo or None
        """

        self.logger.info('Need a repo for %s', self.taginfo['name'])

        # check builds first
        if self.builds:
            # there is no point in requesting a repo if the builds aren't even tagged
            if not koji.util.checkForBuilds(self.session, self.taginfo['id'], self.builds,
                                            event=None):
                self.logger.debug('Builds %s not present in tag %s', self.nvrs,
                                  self.taginfo['name'])
                return None

        check = self.request()
        repoinfo = check.get('repo')
        if repoinfo:
            # "he says they've already got one"
            self.logger.info('Request yielded repo: %r', check)
            if self.check_repo(repoinfo):
                return repoinfo

        # save our request for use in task_args(), the builder code will call that next
        self._task_request = check['request']
        # otherwise
        return None

    def task_args(self):
        """Return args for a waitrepo task matching our data"""
        tag = self.taginfo['name']
        newer_than = None  # this legacy arg doesn't make sense for us
        min_event = self.min_event
        if self.at_event:
            raise koji.GenericError('at_event not supported by waitrepo task')
        if min_event is None:
            # see if we can use value from request
            req = self._task_request
            if req:
                min_event = req['min_event']
        if self.opts:
            # TODO?
            raise koji.GenericError('opts not supported by waitrepo task')
        return [tag, newer_than, self.nvrs, min_event]

    def waitrepo(self, anon=False):
        self.logger.info('Waiting on repo for %s', self.taginfo['name'])
        self.get_start()
        min_event = self.min_event
        self.logger.debug('min_event = %r, nvrs = %r', min_event, self.nvrs)
        repoinfo = None
        req = None
        while True:
            # wait on existing request if we have one
            if req:
                repoinfo = self.wait_request(req)
                if self.check_repo(repoinfo):
                    break
                elif self.at_event is not None:
                    # shouldn't happen
                    raise koji.GenericError('Failed at_event request')
                else:
                    min_event = self.session.tagLastChangeEvent(self.taginfo['id'])
                    # we should have waited for builds before creating the request
                    # this could indicate further tagging/untagging, or a bug
                    self.logger.error('Repo request did not satisfy conditions')
            elif anon:
                # check for repo directly
                repoinfo = self.session.repo.get(self.taginfo['id'], min_event=min_event,
                                                 at_event=self.at_event, opts=self.opts)
                if repoinfo and self.check_repo(repoinfo):
                    break
            # Otherwise, we'll need a new request
            if self.builds:
                # No point in requesting a repo if the builds aren't tagged yet
                self.wait_builds(self.builds)
                min_event = self.session.tagLastChangeEvent(self.taginfo['id'])
                self.logger.debug('Updated min_event to last change: %s', min_event)
            if not anon:
                # Request a repo
                check = self.request(min_event)
                repoinfo = check.get('repo')
                if repoinfo:
                    self.logger.debug('Request yielded repo: %r', check)
                    if self.check_repo(repoinfo):
                        break
                    # otherwise we'll loop and try again
                else:
                    req = check['request']
                    self.logger.info('Got request: %(id)s', req)
                    self.logger.debug('Request data: %s', req)
                    if min_event in ('last', None):
                        min_event = req['min_event']
                        self.logger.info('Updated min_event from hub: %s', min_event)
            self.pause()

        self.logger.debug('Got repo: %r', repoinfo)
        return repoinfo

    def request(self, min_event=None):
        if min_event is None:
            min_event = self.min_event
        self.logger.info('Requesting a repo')
        self.logger.debug('self.session.repo.request(%s, min_event=%s, at_event=%s, opts=%r)',
                          self.taginfo['id'], min_event, self.at_event, self.opts)
        check = self.session.repo.request(self.taginfo['id'], min_event=min_event,
                                          at_event=self.at_event, opts=self.opts)
        return check

    def wait_request(self, req):
        watch_fields = ('task_id', 'task_state', 'repo_id', 'active', 'tries')
        self.get_start()
        watch_data = dict([(f, req.get(f)) for f in watch_fields])
        while True:
            check = self.session.repo.checkRequest(req['id'])
            self.logger.debug('Request check: %r', check)
            repo = check.get('repo')
            if repo:
                return repo
            for f in watch_fields:
                val1 = watch_data[f]
                val2 = check['request'][f]
                if val1 != val2:
                    watch_data[f] = val2
                    if f == 'task_state':
                        # convert if we can
                        val1 = koji.TASK_STATES[val1] if val1 is not None else val1
                        val2 = koji.TASK_STATES[val2] if val2 is not None else val2
                    self.logger.info('Request updated: %s: %s -> %s', f, val1, val2)
            if self.check_timeout():
                raise koji.GenericError("Unsuccessfully waited %s for a new %s repo" %
                                        (koji.util.duration(self.start), self.taginfo['name']))
            if not check['request']['active']:
                if check['request']['task_id']:
                    tstate = koji.TASK_STATES[check['request']['task_state']]
                    msg = 'Repo request inactive: task %s state is %s' % (
                          check['request']['task_id'], tstate)
                else:
                    msg = 'Repo request no longer active'
                raise koji.GenericError(msg)
            self.pause()

    def wait_builds(self, builds):
        self.get_start()
        self.logger.info('Waiting for nvrs %s in tag %s', self.nvrs, self.taginfo['name'])
        while True:
            if koji.util.checkForBuilds(self.session, self.taginfo['id'], builds, event=None):
                self.logger.debug('Successfully waited for nvrs %s in tag %s', self.nvrs,
                                  self.taginfo['name'])
                return
            if self.check_timeout():
                raise koji.GenericError("Unsuccessfully waited %s for %s to appear in the %s repo"
                                        % (koji.util.duration(self.start),
                                           koji.util.printList(self.nvrs),
                                           self.taginfo['name']))
            self.logger.debug('Waiting for nvrs %s in tag %s', self.nvrs, self.taginfo['name'])
            self.pause()

    def check_repo(self, repoinfo):
        """See if the repo satifies our conditions"""

        # Correct tag?
        if repoinfo['tag_id'] != self.taginfo['id']:
            # should not happen
            self.logger.error('Got repo for wrong tag, expected %s, got %s',
                              self.taginfo['id'], repoinfo['tag_id'])
            return False

        # Matching event?
        if self.at_event is not None:
            if repoinfo['create_event'] != self.at_event:
                self.logger.info('Got repo with wrong event. %s != %s',
                                 repoinfo['create_event'], self.at_event)
                return False
        elif self.min_event is not None:
            if repoinfo['create_event'] < self.min_event:
                self.logger.info('Got repo before min event. %s < %s',
                                 repoinfo['create_event'], self.min_event)
                return False

        # Matching opts
        if self.opts is not None:
            if repoinfo['opts'] != self.opts:
                self.logger.info('Got repo with wrong opts. %s != %s',
                                 repoinfo['opts'], self.opts)
                return False

        # Needed builds?
        if self.builds:
            if not koji.util.checkForBuilds(self.session, self.taginfo['id'], self.builds,
                                            event=repoinfo['create_event']):
                self.logger.info('Got repo without needed builds')
                return False

        self.logger.debug('Repo satisfies our conditions')
        return True

    def pause(self):
        self.logger.debug('Pausing')
        time.sleep(self.PAUSE)

    def check_timeout(self):
        if (time.time() - self.start) > (self.TIMEOUT * 60.0):
            return True
        # else
        return False


def duration(start):
    """Return the duration between start and now in MM:SS format"""
    elapsed = time.time() - start
    mins = int(elapsed // 60)
    secs = int(elapsed % 60)
    return '%s:%02i' % (mins, secs)


def printList(lst):
    """Print the contents of the list comma-separated"""
    if len(lst) == 0:
        return ''
    elif len(lst) == 1:
        return lst[0]
    elif len(lst) == 2:
        return ' and '.join(lst)
    else:
        ret = ', '.join(lst[:-1])
        ret += ', and '
        ret += lst[-1]
        return ret


def base64encode(s, as_bytes=False):
    """Helper function to encode string or bytes as base64

    This function returns a string unless as_bytes is True
    """
    if six.PY2:
        return base64.b64encode(s)

    if isinstance(s, str):
        s = s.encode('utf8')
    data = base64.b64encode(s)
    if as_bytes:
        return data
    else:
        # ascii is always good enough for base64 encoded data
        return data.decode('ascii')


# We don't need a decode wrapper, but we define this for naming consistency
base64decode = base64.b64decode


def decode_bytes(data, fallback='iso8859-15'):
    """Decode a bytes-like object that is expected to be a valid string

    First utf8 is tried, then the fallback (defaults to iso8859-15).
    The fallback behavior can be disabled by setting the option to None.
    """
    try:
        return data.decode('utf8')
    except UnicodeDecodeError:
        if fallback:
            return data.decode(fallback)
        raise


def multi_fnmatch(s, patterns):
    """Returns true if s matches any pattern in the list

    If patterns is a string, it will be split() first
    """
    if isinstance(patterns, six.string_types):
        patterns = patterns.split()
    for pat in patterns:
        if fnmatch(s, pat):
            return True
    return False


def dslice(dict_, keys, strict=True):
    """Returns a new dictionary containing only the specified keys"""
    ret = {}
    for key in keys:
        if strict or key in dict_:
            # for strict we skip the has_key check and let the dict generate the KeyError
            ret[key] = dict_[key]
    return ret


def dslice_ex(dict_, keys, strict=True):
    """Returns a new dictionary with only the specified keys removed"""
    ret = dict_.copy()
    for key in keys:
        if strict or key in ret:
            del ret[key]
    return ret


class DataWalker(object):

    def __init__(self, data, callback, kwargs=None):
        self.data = data
        self.callback = callback
        if kwargs is None:
            kwargs = {}
        self.kwargs = kwargs

    def walk(self):
        return self._walk(self.data)

    def _walk(self, value):
        # recurse if needed
        if isinstance(value, tuple):
            value = tuple([self._walk(x) for x in value])
        elif isinstance(value, list):
            value = [self._walk(x) for x in value]
        elif isinstance(value, dict):
            ret = {}
            for k in value:
                k = self._walk(k)
                v = self._walk(value[k])
                ret[k] = v
            value = ret
        # finally, let callback filter the value
        return self.callback(value, **self.kwargs)


def encode_datetime(value):
    """Convert datetime objects to strings"""
    if isinstance(value, datetime.datetime):
        return value.isoformat(' ')
    elif isinstance(value, DateTime):
        return datetime.datetime(*value.timetuple()[:6]).isoformat(' ')
    else:
        return value


def encode_datetime_recurse(value):
    walker = DataWalker(value, encode_datetime)
    return walker.walk()


def call_with_argcheck(func, args, kwargs=None):
    """Call function, raising ParameterError if args do not match"""
    if kwargs is None:
        kwargs = {}
    try:
        return func(*args, **kwargs)
    except TypeError as e:
        if sys.exc_info()[2].tb_next is None:
            # The stack is only one high, so the error occurred in this function.
            # Therefore, we assume the TypeError is due to a parameter mismatch
            # in the above function call.
            raise koji.ParameterError(str(e))
        raise


def apply_argspec(argspec, args, kwargs=None):
    """Apply an argspec to the given args and return a dictionary"""
    if kwargs is None:
        kwargs = {}
    f_args, f_varargs, f_varkw, f_defaults = argspec
    data = dict(zip(f_args, args))
    if len(args) > len(f_args):
        if not f_varargs:
            raise koji.ParameterError('too many args')
        data[f_varargs] = tuple(args[len(f_args):])
    elif f_varargs:
        data[f_varargs] = ()
    if f_varkw:
        data[f_varkw] = {}
    for arg in kwargs:
        if arg in data:
            raise koji.ParameterError('duplicate keyword argument %r' % arg)
        if arg in f_args:
            data[arg] = kwargs[arg]
        elif not f_varkw:
            raise koji.ParameterError("unexpected keyword argument %r" % arg)
        else:
            data[f_varkw][arg] = kwargs[arg]
    if f_defaults:
        for arg, val in zip(f_args[-len(f_defaults):], f_defaults):
            data.setdefault(arg, val)
    for n, arg in enumerate(f_args):
        if arg not in data:
            raise koji.ParameterError('missing required argument %r (#%i)'
                                      % (arg, n))
    return data


class HiddenValue(object):
    """A wrapper that prevents a value being accidentally printed"""

    def __init__(self, value):
        if isinstance(value, HiddenValue):
            self.value = value.value
        else:
            self.value = value

    def __str__(self):
        return "[value hidden]"

    def __repr__(self):
        return "HiddenValue()"


class LazyValue(object):
    """Used to represent a value that is generated by a function call at access time
    """

    def __init__(self, func, args, kwargs=None, cache=False):
        if kwargs is None:
            kwargs = {}
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.cache = cache

    def get(self):
        if hasattr(self, '_value'):
            return self._value
        value = self.func(*self.args, **self.kwargs)
        if self.cache:
            self._value = value
        return value


class LazyString(LazyValue):
    """Lazy values that should be expanded when printed"""

    def __str__(self):
        return str(self.get())


def lazy_eval(value):
    if isinstance(value, LazyValue):
        return value.get()
    return value


class LazyDict(dict):
    """A container for lazy data

    fields can refer to function calls, which can optionally be cached
    """

    def __getitem__(self, key):
        return lazy_eval(super(LazyDict, self).__getitem__(key))

    def lazyset(self, key, func, args, kwargs=None, cache=False):
        self[key] = LazyValue(func, args, kwargs=kwargs, cache=cache)

    def get(self, *args, **kwargs):
        return lazy_eval(super(LazyDict, self).get(*args, **kwargs))

    def copy(self):
        return LazyDict(self)

    def values(self):
        return [lazy_eval(val) for val in super(LazyDict, self).values()]

    def items(self):
        return [(key, lazy_eval(val)) for key, val in super(LazyDict, self).items()]

    def itervalues(self):
        for val in six.itervalues(super(LazyDict, self)):
            yield lazy_eval(val)

    def iteritems(self):
        for key, val in six.iteritems(super(LazyDict, self)):
            yield key, lazy_eval(val)

    def pop(self, key, *args, **kwargs):
        return lazy_eval(super(LazyDict, self).pop(key, *args, **kwargs))

    def popitem(self):
        key, val = super(LazyDict, self).popitem()
        return key, lazy_eval(val)


class LazyRecord(object):
    """A object whose attributes can reference lazy data

    Use lazysetattr to set lazy attributes, or just set them to a LazyValue
    object directly"""

    def __init__(self, base=None):
        if base is not None:
            self.__dict__.update(base.__dict__)
        self._base_record = base

    def __getattribute__(self, name):
        try:
            val = object.__getattribute__(self, name)
        except AttributeError:
            base = object.__getattribute__(self, '_base_record')
            val = getattr(base, name)
        return lazy_eval(val)


def lazysetattr(object, name, func, args, kwargs=None, cache=False):
    if not isinstance(object, LazyRecord):
        raise TypeError('object does not support lazy attributes')
    value = LazyValue(func, args, kwargs=kwargs, cache=cache)
    setattr(object, name, value)


class _RetryRmtree(Exception):
    """This exception is used internally by rmtree"""
    # We raise this exception only when it makes sense for rmtree to retry from the top


def rmtree(path, logger=None, background=False):
    """Delete a directory tree without crossing fs boundaries

    :param str path: the directory to remove
    :param Logger logger: Logger object
    :param bool background: if True, runs in the background returning a cleanup function
    :return: None or [pid, check_function]

    In the background case, caller is responsible for waiting on pid and should call
    the returned check function once it finishes.
    """
    # we use the fake logger to avoid issues with logging locks while forking
    fd, logfile = tempfile.mkstemp(suffix='.jsonl')
    os.close(fd)
    pid = os.fork()

    if not pid:
        # child process
        try:
            status = 1
            with SimpleProxyLogger(logfile) as mylogger:
                try:
                    _rmtree_nofork(path, logger=mylogger)
                except Exception as e:
                    mylogger.error('rmtree failed: %s' % e)
                    raise
            status = 0
        finally:
            # diediedie
            os._exit(status)
        # not reached

    # parent process
    logger = logger or logging.getLogger('koji')

    def _rmtree_check():
        if not background:
            # caller will wait in background case
            _pid, status = os.waitpid(pid, 0)
        try:
            SimpleProxyLogger.send(logfile, logger)
        except Exception as err:
            logger.error("Failed to get rmtree logs -- %s" % err)
        try:
            os.unlink(logfile)
        except Exception:
            pass
        if not background:
            # caller should check status in background case
            if not isSuccess(status):
                raise koji.GenericError(parseStatus(status, "rmtree process"))
        if os.path.exists(path):
            raise koji.GenericError("Failed to remove directory: %s" % path)

    if background:
        return pid, _rmtree_check
    else:
        return _rmtree_check()


class SimpleProxyLogger(object):
    """Save log messages to a file and log them later"""

    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR

    def __init__(self, filename):
        self.outfile = koji._open_text_file(filename, mode='wt')

    # so we can use as a context manager
    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.outfile.close()
        # don't eat exceptions
        return False

    def log(self, level, msg, *args, **kwargs):
        # jsonl output
        data = [level, msg, args, kwargs]
        try:
            line = json.dumps(data, indent=None)
        except Exception:
            try:
                data = [logging.ERROR, "Unable to log: %s" % data, (), {}]
                line = json.dumps(data, indent=None)
            except Exception:
                line = '[40, "Invalid log data", [], {}]'
        try:
            self.outfile.write(line)
            self.outfile.write('\n')
        except Exception:
            pass

    def info(self, msg, *args, **kwargs):
        self.log(self.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(self.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log(self.ERROR, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.log(self.DEBUG, msg, *args, **kwargs)

    @staticmethod
    def send(filename, logger):
        with koji._open_text_file(filename, mode='rt') as fo:
            for line in fo:
                try:
                    level, msg, args, kwargs = json.loads(line)
                except Exception:
                    level = logging.ERROR
                    msg = "Bad log data: %r"
                    args = (line,)
                    kwargs = {}
                logger.log(level, msg, *args, **kwargs)


def _rmtree_nofork(path, logger=None):
    """Delete a directory tree without crossing fs boundaries

    This function is not thread safe because it relies on chdir to avoid
    forming long paths.
    """
    # implemented to avoid forming long paths
    # see: https://pagure.io/koji/issue/201
    logger = logger or logging.getLogger('koji')
    try:
        st = os.lstat(path)
    except OSError as e:
        # FileNotFoundError is N/A in py2
        if e.errno == errno.ENOENT:
            logger.warning("No such file/dir %s for removal" % path)
            return
        raise
    if not stat.S_ISDIR(st.st_mode):
        raise koji.GenericError("Not a directory: %s" % path)
    dev = st.st_dev
    cwd = os.getcwd()
    abspath = os.path.abspath(path)

    try:
        # retry loop
        while True:
            try:
                os.chdir(path)
                new_cwd = os.getcwd()
                # make sure we're where we think we are
                if not os.path.samefile(new_cwd, abspath):
                    raise koji.GenericError('chdir to %s resulted in different cwd %s',
                                            path, new_cwd)
            except OSError as e:
                if e.errno in (errno.ENOENT, errno.ESTALE):
                    # likely racing with another rmtree
                    # if the dir doesn't exist, we're done
                    logger.warning("Directory to remove has disappeared: %s" % path)
                    return
                raise
            try:
                _rmtree(dev, new_cwd, logger)
            except _RetryRmtree as e:
                # reset and retry
                os.chdir(cwd)
                logger.warning("Retrying rmtree due to %s" % e)
                continue
            break
    finally:
        os.chdir(cwd)

    # a successful _rmtree call should leave us with an empty directory
    try:
        os.rmdir(path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def _rmtree(dev, cwd, logger):
    """Remove all contents of CWD"""
    # This implementation avoids forming long paths and recursion. Otherwise
    # we will have errors with very deep directory trees.
    # - to avoid forming long paths we change directory as we go
    # - to avoid recursion we maintain our own stack
    dirstack = []
    # Each entry in dirstack contains data for a level of directory traversal
    #    - path
    #    - subdirs
    # As we descend into the tree, we append a new entry to dirstack
    # When we ascend back up after removal, we pop them off

    while True:
        dirs = _stripcwd(dev, cwd, logger)

        # if cwd has no subdirs, walk back up until we find some
        while not dirs and dirstack:
            _assert_cwd(cwd)
            try:
                os.chdir('..')
            except OSError as e:
                _assert_cwd(cwd)
                if e.errno in (errno.ENOENT, errno.ESTALE):
                    # likely in a race with another rmtree
                    # however, we cannot proceed from here, so we return to the top
                    raise _RetryRmtree(str(e))
                raise
            dirs = dirstack.pop()
            cwd = os.path.dirname(cwd)

            # now that we've ascended back up by one, the last dir entry is
            # one we've just cleared, so we should remove it
            empty_dir = dirs.pop()
            _assert_cwd(cwd)
            try:
                os.rmdir(empty_dir)
            except OSError as e:
                _assert_cwd(cwd)
                # If this happens, either something else is writing to the dir,
                # or there is a bug in our code.
                # For now, we ignore this and proceed, but we'll still fail at
                # the top level rmdir
                logger.error("Unable to remove directory %s: %s" % (empty_dir, e))
                pass

        if not dirs:
            # we are done
            break

        # otherwise we descend into the next subdir
        subdir = dirs[-1]
        # note: we do not pop here because we need to remember to remove subdir later
        _assert_cwd(cwd)
        try:
            os.chdir(subdir)
        except OSError as e:
            _assert_cwd(cwd)
            if e.errno == errno.ENOENT:
                # likely in a race with another rmtree
                # we'll ignore this and continue
                # since subdir doesn't exist, we'll pop it off and forget about it
                dirs.pop()
                logger.warning("Subdir disappeared during rmtree %s: %s" % (subdir, e))
                continue  # with dirstack unchanged
            raise
        cwd = os.path.join(cwd, subdir)
        dirstack.append(dirs)


def _assert_cwd(cwd):
    try:
        actual = os.getcwd()
    except OSError as e:
        if e.errno == errno.ENOENT:
            # subsequent calls should fail with better handling
            return
        raise
    if cwd != actual:
        raise koji.GenericError('CWD changed unexpectedly: %s -> %s' % (cwd, actual))


def _stripcwd(dev, cwd, logger):
    """Unlink all files in cwd and return list of subdirs"""
    dirs = []
    _assert_cwd(cwd)
    try:
        fdirs = os.listdir('.')
    except OSError as e:
        if e.errno in (errno.ENOENT, errno.ESTALE):
            # cwd could have been removed by others, just return an empty list
            logger.warning("Unable to read directory: %s" % e)
            return dirs
        raise
    for fn in fdirs:
        try:
            st = os.lstat(fn)
        except OSError as e:
            _assert_cwd(cwd)
            if e.errno == errno.ENOENT:
                continue
            raise
        if st.st_dev != dev:
            # don't cross fs boundary
            continue
        if stat.S_ISDIR(st.st_mode):
            dirs.append(fn)
        else:
            _assert_cwd(cwd)
            try:
                os.unlink(fn)
            except OSError:
                # we'll still fail at the top level
                pass
    return dirs


def safer_move(src, dst):
    """Rename if possible, copy+rm otherwise

    Behavior is similar to shutil.move

    Unlike move, src is /always/ moved from src to dst. If dst is an existing
    directory, then an error is raised.
    """
    if os.path.exists(dst):
        raise koji.GenericError("Destination exists: %s" % dst)
    elif os.path.islink(dst):
        raise koji.GenericError("Destination is a symlink: %s" % dst)
    # TODO - use locking to do a better job of catching races
    shutil.move(src, dst)


def move_and_symlink(src, dst, relative=True, create_dir=False):
    """Move src to dest and create symlink instead of original file"""
    if create_dir:
        koji.ensuredir(os.path.dirname(dst))
    safer_move(src, dst)
    if relative:
        dst = os.path.relpath(dst, os.path.dirname(src))
    os.symlink(dst, src)


def joinpath(path, *paths):
    """A wrapper around os.path.join that limits directory traversal"""

    # note that the first path is left alone

    newpaths = []
    for _p in paths:
        p = os.path.normpath(_p)
        if p == '..' or p.startswith('../') or p.startswith('/'):
            raise ValueError('Invalid path segment: %s' % _p)
        newpaths.append(p)

    return os.path.join(path, *newpaths)


def eventFromOpts(session, opts):
    """Determine event id from standard cli options

    Standard options are:
        event: an event id (int)
        ts: an event timestamp (int)
        repo: pull event from given repo
    """
    event_id = getattr(opts, 'event', None)
    if event_id:
        return session.getEvent(event_id)
    ts = getattr(opts, 'ts', None)
    if ts is not None:
        return session.getLastEvent(before=ts)
    repo = getattr(opts, 'repo', None)
    if repo is not None:
        rinfo = session.repoInfo(repo, strict=True)
        return {'id': rinfo['create_event'],
                'ts': rinfo['create_ts']}
    return None


def filedigestAlgo(hdr):
    """
    Get the file digest algorithm used in hdr.
    If there is no algorithm flag in the header,
    default to md5.  If the flag contains an unknown,
    non-None value, return 'unknown'.
    """
    # need to use the header ID hard-coded into Koji so we're not dependent on the
    # version of rpm installed on the hub
    digest_algo_id = hdr[koji.RPM_TAG_FILEDIGESTALGO]
    if not digest_algo_id:
        # certain versions of rpm return an empty list instead of None
        # for missing header fields
        digest_algo_id = None
    digest_algo = koji.RPM_FILEDIGESTALGO_IDS.get(digest_algo_id, 'unknown')
    return digest_algo.lower()


def check_sigmd5(filename):
    """Compare header's sigmd5 with actual md5 of hdr+payload without need of rpm"""
    with open(filename, 'rb') as f:
        leadsize = 96
        # skip magic + reserved
        o = leadsize + 8
        f.seek(o)
        data = f.read(8)
        indexcount, storesize = struct.unpack('!II', data)
        for idx in range(indexcount):
            data = f.read(16)
            tag, data_type, offset, count = struct.unpack('!IIII', data)
            if tag == 1004:  # SIGMD5
                assert (data_type == 7)  # binary data
                assert (count == 16)     # 16 bytes of md5
                break
        else:
            # rpm v6 format does not include this header
            return True
        # seek to location of md5
        f.seek(o + 8 + indexcount * 16 + offset)
        sigmd5 = f.read(16)

        # seek to start of header
        sigsize = 8 + 16 * indexcount + storesize
        o += sigsize + (8 - (sigsize % 8)) % 8
        f.seek(o)

        # compute md5 of rest of file
        md5 = md5_constructor()
        while True:
            d = f.read(1024**2)
            if not d:
                break
            md5.update(d)

        return sigmd5 == md5.digest()


def parseStatus(rv, prefix):
    if isinstance(prefix, (list, tuple)):
        prefix = ' '.join(prefix)
    if os.WIFSIGNALED(rv):
        return '%s was killed by signal %i' % (prefix, os.WTERMSIG(rv))
    elif os.WIFEXITED(rv):
        return '%s exited with status %i' % (prefix, os.WEXITSTATUS(rv))
    else:
        return '%s terminated for unknown reasons' % prefix


def isSuccess(rv):
    """Return True if rv indicates successful completion
    (exited with status 0), False otherwise."""
    if os.WIFEXITED(rv) and os.WEXITSTATUS(rv) == 0:
        return True
    else:
        return False


def setup_rlimits(opts, logger=None):
    logger = logger or logging.getLogger("koji")
    for key in opts:
        if not key.startswith('RLIMIT_') or not opts[key]:
            continue
        rcode = getattr(resource, key, None)
        if rcode is None:
            continue
        orig = resource.getrlimit(rcode)
        try:
            limits = [int(x) for x in opts[key].split()]
        except ValueError:
            logger.error("Invalid resource limit: %s=%s", key, opts[key])
            continue
        if len(limits) not in (1, 2):
            logger.error("Invalid resource limit: %s=%s", key, opts[key])
            continue
        if len(limits) == 1:
            limits.append(orig[1])
        logger.warning('Setting resource limit: %s = %r', key, limits)
        try:
            resource.setrlimit(rcode, tuple(limits))
        except ValueError as e:
            logger.error("Unable to set %s: %s", key, e)


class adler32_constructor(object):

    # mimicing the hashlib constructors
    def __init__(self, arg=''):
        if six.PY3 and isinstance(arg, str):
            arg = bytes(arg, 'utf-8')
        self._value = adler32(arg) & 0xffffffff
        # the bitwise and works around a bug in some versions of python
        # see: https://bugs.python.org/issue1202

    def update(self, arg):
        if six.PY3 and isinstance(arg, str):
            arg = bytes(arg, 'utf-8')
        self._value = adler32(arg, self._value) & 0xffffffff

    def digest(self):
        return self._value

    def hexdigest(self):
        return "%08x" % self._value

    def copy(self):
        dup = adler32_constructor()
        dup._value = self._value
        return dup

    digest_size = 4
    block_size = 1  # I think


def tsort(parts):
    """Given a partial ordering, return a totally ordered list.

    part is a dict of partial orderings.  Each value is a set,
    which the key depends on.

    The return value is a list of sets, each of which has only
    dependencies on items in previous entries in the list."""
    parts = parts.copy()
    result = []
    while True:
        level = set([name for name, deps in six.iteritems(parts) if not deps])
        if not level:
            break
        result.append(level)
        parts = dict([(name, deps - level) for name, deps in six.iteritems(parts)
                      if name not in level])
    if parts:
        raise ValueError('total ordering not possible')
    return result


class MavenConfigOptAdapter(object):
    """
    Wrap a ConfigParser so it looks like a optparse.Values instance
    used by maven-build.
    """
    MULTILINE = ['properties', 'envs']
    MULTIVALUE = ['goals', 'profiles', 'packages',
                  'jvm_options', 'maven_options', 'buildrequires']

    def __init__(self, conf, section):
        self._conf = conf
        self._section = section

    def __getattr__(self, name):
        if self._conf.has_option(self._section, name):
            value = self._conf.get(self._section, name)
            if name in self.MULTIVALUE:
                value = value.split()
            elif name in self.MULTILINE:
                value = value.splitlines()
            return value
        raise AttributeError(name)


def maven_opts(values, chain=False, scratch=False):
    """
    Convert the argument (an optparse.Values object) to a dict of build options
    suitable for passing to maven-build or maven-chain.
    """
    opts = {}
    for key in ('scmurl', 'patches', 'specfile', 'goals', 'profiles', 'packages',
                'jvm_options', 'maven_options'):
        val = getattr(values, key, None)
        if val:
            opts[key] = val
    props = {}
    for prop in getattr(values, 'properties', []):
        fields = prop.split('=', 1)
        if len(fields) != 2:
            fields.append(None)
        props[fields[0]] = fields[1]
    if props:
        opts['properties'] = props
    envs = {}
    for env in getattr(values, 'envs', []):
        fields = env.split('=', 1)
        if len(fields) != 2:
            raise ValueError("Environment variables must be in NAME=VALUE format")
        envs[fields[0]] = fields[1]
    if envs:
        opts['envs'] = envs
    if chain:
        val = getattr(values, 'buildrequires', [])
        if val:
            opts['buildrequires'] = val
    if scratch and not chain:
        opts['scratch'] = True
    return opts


def maven_params(config, package, chain=False, scratch=False):
    values = MavenConfigOptAdapter(config, package)
    return maven_opts(values, chain=chain, scratch=scratch)


def wrapper_params(config, package, chain=False, scratch=False):
    params = {}
    values = MavenConfigOptAdapter(config, package)
    params['type'] = getattr(values, 'type', None)
    params['scmurl'] = getattr(values, 'scmurl', None)
    params['buildrequires'] = getattr(values, 'buildrequires', [])
    if not scratch:
        params['create_build'] = True
    return params


def parse_maven_params(confs, chain=False, scratch=False):
    """
    Parse .ini files that contain parameters to launch a Maven build.

    Return a map whose keys are package names and values are config parameters.
    """
    config = koji.read_config_files(confs)
    builds = {}
    for package in config.sections():
        buildtype = 'maven'
        if config.has_option(package, 'type'):
            buildtype = config.get(package, 'type')
        if buildtype == 'maven':
            params = maven_params(config, package, chain=chain, scratch=scratch)
        elif buildtype == 'wrapper':
            params = wrapper_params(config, package, chain=chain, scratch=scratch)
            if len(params.get('buildrequires')) != 1:
                raise ValueError("A wrapper-rpm must depend on exactly one package")
        else:
            raise ValueError("Unsupported build type: %s" % buildtype)
        if 'scmurl' not in params:
            raise ValueError("%s is missing the scmurl parameter" % package)
        builds[package] = params
    if not builds:
        if not isinstance(confs, (list, tuple)):
            confs = [confs]
        raise ValueError("No sections found in: %s" % ', '.join(confs))
    return builds


def parse_maven_param(confs, chain=False, scratch=False, section=None):
    """
    Parse .ini files that contain parameters to launch a Maven build.

    Return a map that contains a single entry corresponding to the given
    section of the .ini file.  If the config file only contains a single
    section, section does not need to be specified.
    """
    if not isinstance(confs, (list, tuple)):
        confs = [confs]
    builds = parse_maven_params(confs, chain=chain, scratch=scratch)
    if section:
        if section in builds:
            builds = {section: builds[section]}
        else:
            raise ValueError("Section %s does not exist in: %s" % (section, ', '.join(confs)))
    elif len(builds) > 1:
        raise ValueError(
            "Multiple sections in: %s, you must specify the section" % ', '.join(confs))
    return builds


def parse_maven_chain(confs, scratch=False):
    """
    Parse maven-chain config.

    confs is a path to a config file or a list of paths to config files.

    Return a map whose keys are package names and values are config parameters.
    """
    builds = parse_maven_params(confs, chain=True, scratch=scratch)
    depmap = {}
    for package, params in builds.items():
        depmap[package] = set(params.get('buildrequires', []))
    try:
        tsort(depmap)
    except ValueError:
        raise ValueError('No possible build order, missing/circular dependencies')
    return builds


def to_list(lst):
    """
    Helper function for py2/py3 compatibility used e.g. in
    list(dict.keys())

    Don't use it for structures like list(zip(x, y)), where six.moves.zip is
    used, so it is always an iterator.
    """

    if isinstance(lst, list):
        return lst
    else:
        return list(lst)


def format_shell_cmd(cmd, text_width=80):
    """
    Helper for wrapping shell command lists to human-readable form, while
    they still can be copied (from logs) and run in shell.

    :param [str] cmd: command list
    :returns str:
    """
    # account for " \"
    text_width -= 2
    s = []
    line = ''
    for bit in cmd:
        if len(line + bit) > text_width:
            if line:
                s.append(line)
                line = ''
        if line:
            line += ' '
        line += bit
    if line:
        s.append(line)
    return ' \\\n'.join(s)


def extract_build_task(binfo):
    """
    Helper for extracting task id from buildinfo. CGs and older OSBS approach
    can put it into different places in binfo

    :param dict binfo: buildinfo
    :returns int: task id
    """

    task_id = binfo.get('task_id')
    if task_id is None:
        # legacy OSBS task id location
        extra = binfo.get('extra')
        if extra is not None:
            task_id = extra.get('container_koji_task_id')
    return task_id
